from typing import List, Optional, Dict, Any

from lucyfer.searchset.mapping import Mapping
from lucyfer.searchset.utils import FieldType
from lucyfer.utils import fill_field_if_it_necessary


# todo merge those classes

class SearchHelper:
    """
    SearchHelperMixin provides possibility to get mapping and search helpers for user-friendly search API
    """

    fields_to_exclude_from_mapping: Optional[List[str]] = None
    fields_to_exclude_from_suggestions: Optional[List[str]] = None
    show_suggestions = True
    escape_quotes_in_suggestions = True

    _mapping_class = None

    _full_mapping: Optional[Mapping] = None
    _raw_mapping = None

    @classmethod
    def get_mapping(cls) -> List[str]:
        """
        Returns full mapping with handwritten fields in search set class and their sources
        Except of excluded fields/sources
        """
        return list(cls.get_full_mapping().keys())

    @classmethod
    def get_mapping_to_suggestion(cls) -> Dict[str, bool]:
        return {k: v.show_suggestions for k, v in cls.get_full_mapping().items()}

    @classmethod
    def get_fields_values(cls, qs, field_name, prefix='', cache_key=None) -> List[str]:
        """
        Returns search helpers for field by prefix
        """
        return cls.get_full_mapping().get_values(qs=qs, field_name=field_name, prefix=prefix, cache_key=cache_key)

    @classmethod
    def get_fields_to_exclude_from_mapping(cls) -> List[str]:
        return cls.fields_to_exclude_from_mapping if cls.fields_to_exclude_from_mapping is not None else list()

    @classmethod
    def get_fields_to_exclude_from_suggestions(cls) -> List[str]:
        return cls.fields_to_exclude_from_suggestions if cls.fields_to_exclude_from_suggestions is not None else list()

    @classmethod
    def get_raw_mapping(cls) -> Dict[str, FieldType]:
        """
        Caches raw mapping and return it
        """
        if cls._raw_mapping is None:
            cls._raw_mapping = cls._get_raw_mapping()
        return cls._raw_mapping

    @classmethod
    def get_full_mapping(cls):
        if cls._full_mapping is None:
            cls._fill_mapping()
        return cls._full_mapping

    @classmethod
    def _fill_mapping(cls):
        """
        Fill mapping extended by handwritten fields and its sources
        """

        mapping_exclude = cls.get_fields_to_exclude_from_mapping()
        suggestions_exclude = cls.get_fields_to_exclude_from_suggestions()

        mapping = cls._mapping_class(cls.Meta.model)

        # create mapping values from fields in searchset class
        for field_name, field in cls.get_field_name_to_field().items():

            get_available_values_method = field.get_available_values_method()

            if field_name not in mapping_exclude:
                mapping.add_value(name=field_name,
                                  sources=field.sources,
                                  show_suggestions=(cls.show_suggestions and field.show_suggestions
                                                    and field_name not in suggestions_exclude),
                                  get_available_values_method=get_available_values_method,
                                  escape_quotes_in_suggestions=cls.escape_quotes_in_suggestions)

            if not field.exclude_sources_from_mapping:
                if field.use_field_class_for_sources:
                    for source in field.sources:
                        mapping.add_value(name=source,
                                          show_suggestions=cls.show_suggestions and source not in suggestions_exclude,
                                          get_available_values_method=get_available_values_method,
                                          escape_quotes_in_suggestions=cls.escape_quotes_in_suggestions)
            else:
                mapping_exclude.extend(field.sources)

        # update mapping from mapping in database/elastic/etc
        raw_mapping = cls.get_raw_mapping()

        for name, field_type in raw_mapping.items():
            if name not in mapping_exclude:
                if field_type in cls._field_type_to_field_class:
                    field_instance = cls._field_type_to_field_class[field_type]()
                else:
                    field_instance = None

                get_available_values_method = field_instance.get_available_values_method() if field_instance else None
                mapping.add_value(name=name,
                                  field_type=field_type,
                                  show_suggestions=cls.show_suggestions and name not in suggestions_exclude,
                                  get_available_values_method=get_available_values_method,
                                  escape_quotes_in_suggestions=cls.escape_quotes_in_suggestions)

                if field_instance is not None:
                    cls.append_field_source_to_search_field_instance(source=name,
                                                                     instance=field_instance)

        cls._full_mapping = mapping

    @classmethod
    def _get_raw_mapping(cls) -> Dict[str, FieldType]:
        """
        That method allows to get mapping in raw format. It have to be reimplemented
        """
        raise NotImplementedError()


class BaseSearchSet(SearchHelper):
    _field_base_class = None

    _field_name_to_search_field_instance: Optional[Dict[str, _field_base_class]] = None
    _field_source_to_search_field_instance: Optional[Dict[str, _field_base_class]] = None

    # default field uses for creating query for fields not defined in searchset class
    _default_field = None

    # provides possibility to use auto cast for boolean/integer/etc fields by field classes usage
    # that means we analyze elastic mapping data types or django models to match it to field classes
    _field_type_to_field_class: Optional[Dict[int, _field_base_class]] = None
    _raw_type_to_field_type: Optional[Dict[Any, int]] = None

    @classmethod
    def get_field_name_to_field(cls):
        """
        Returns dictionary with fieldnames defined in searchset class and field instances
        """
        if cls._field_name_to_search_field_instance is None:
            cls._field_name_to_search_field_instance = dict()
            cls._field_source_to_search_field_instance = dict()

            for name, _cls in cls.__dict__.items():
                if isinstance(_cls, cls._field_base_class):

                    cls._field_name_to_search_field_instance[name] = _cls

                    if _cls.use_field_class_for_sources and not _cls.exclude_sources_from_mapping and _cls.sources:
                        for field_source in _cls.sources:
                            cls.append_field_source_to_search_field_instance(source=field_source, instance=_cls.__class__())

        return cls._field_name_to_search_field_instance

    @classmethod
    @fill_field_if_it_necessary({"_field_source_to_search_field_instance": dict})
    def append_field_source_to_search_field_instance(cls, source, instance):
        cls._field_source_to_search_field_instance[source] = instance

    @classmethod
    def get_query_for_field(cls, condition):
        """
        Returns Q object with query for parsed condition
        """

        field = cls.get_field_name_to_field().get(condition.name)

        # if field not presented in hardcoded fields in searchset class maybe it presented in some field sources
        if field is None:
            field = cls._field_source_to_search_field_instance.get(condition.name)

        # if fields is not source we can try to get field instance by field type from mapping
        if field is None:

            # check field type
            field_type = getattr(cls.get_full_mapping().get(condition.name), "field_type", None)

            field = cls._field_type_to_field_class.get(field_type, cls._default_field)()

        # and finally create a query
        return field.get_query(condition)



from django.db.models import Q
from lucyparser.tree import Operator

from lucyfer.searchset.fields.base import BaseSearchField, negate_query_if_necessary
from lucyfer.searchset.utils import FieldType
from lucyfer.utils import LuceneSearchCastValueException


class DjangoSearchFieldWithoutWildcard(BaseSearchField):
    DEFAULT_LOOKUP = "icontains"

    def create_query_for_sources(self, condition):
        query = Q()

        lookup = self.get_lookup(condition.operator)
        value = self.cast_value(condition.value)

        for source in self.get_sources(condition.name):
            query = query | Q(**{"{}__{}".format(source, lookup): value})
        return query

    @negate_query_if_necessary
    def get_query(self, condition):
        if self.match_all(value=condition.value):
            return None

        return self.create_query_for_sources(condition=condition)


class DjangoSearchField(DjangoSearchFieldWithoutWildcard):
    DEFAULT_LOOKUP = "iexact"

    def process_wildcard(self, value):
        if value.startswith("*") and value.endswith("*"):
            return value[1:-1], "icontains"
        elif value.startswith("*"):
            return value[1:], "iendswith"
        elif value.endswith("*"):
            return value[:-1], "istartswith"

        return value, None

    def create_query_for_sources(self, condition):
        value = self.cast_value(condition.value)

        value, lookup = self.process_wildcard(value=value)

        if lookup is None:
            return super().create_query_for_sources(condition)

        if not value:
            return None

        sources = self.get_sources(condition.name)

        query = Q()

        for source in sources:
            query = query | Q(**{"{}__{}".format(source, lookup): value})

        return query


class DjangoCharField(DjangoSearchField):
    OPERATOR_TO_LOOKUP = {
        Operator.EQ: "icontains",
        Operator.NEQ: "iexact",
    }


class DjangoNumberField(DjangoSearchFieldWithoutWildcard):
    OPERATOR_TO_LOOKUP = {
        Operator.GTE: "gte",
        Operator.LTE: "lte",
        Operator.GT: "gt",
        Operator.LT: "lt",
        Operator.EQ: "exact",
        Operator.NEQ: "exact",
    }


class DjangoIntegerField(DjangoNumberField):
    def cast_value(self, value):
        try:
            return int(value)
        except (ValueError, TypeError):
            raise LuceneSearchCastValueException()


class DjangoFloatField(DjangoNumberField):
    def cast_value(self, value):
        try:
            return float(value)
        except (ValueError, TypeError):
            raise LuceneSearchCastValueException()


class DjangoBooleanField(DjangoSearchFieldWithoutWildcard):
    OPERATOR_TO_LOOKUP = {
        Operator.EQ: "exact",
        Operator.NEQ: "exact",
    }

    _values = {"true": True, "false": False}
    _default_get_available_values_method = _values.keys

    def cast_value(self, value):
        value = value.lower()
        if value in self._values:
            return self._values[value]

        raise LuceneSearchCastValueException()


class DjangoNullBooleanField(DjangoBooleanField):
    _values = {"true": True, "false": False, "null": None}
    _default_get_available_values_method = _values.keys


default_django_field_types_to_fields = {
    FieldType.INTEGER: DjangoIntegerField,
    FieldType.BOOLEAN: DjangoBooleanField,
    FieldType.NULL_BOOLEAN: DjangoNullBooleanField,
    FieldType.FLOAT: DjangoFloatField
}



from collections import defaultdict
from typing import List

from lucyfer.searchset.mapping.base import MappingValue, Mapping


class ElasticMappingValue(MappingValue):
    def _get_values(self, qs, prefix) -> List[str]:
        search = qs.extra(size=0).query('query_string', **{"fields": self.sources, "query": f'*{prefix}*'})
        for source in self.sources:
            search.aggs.bucket(source, "terms", field=source)

        aggs = search.execute().to_dict().get("aggregations", defaultdict(dict))
        result = []
        for source in self.sources:
            if aggs[source].get("buckets"):
                result.extend(
                    [val for val
                     in [str(val["key"]) for val in aggs[source]["buckets"]]
                     if prefix in val]
                )

        return list(set(result))


class ElasticMapping(Mapping):
    _value_class = ElasticMappingValue



from rest_framework.compat import distinct
from rest_framework.filters import SearchFilter

from lucyfer.utils import LuceneSearchException


class DjangoLuceneSearchFilterMixin(SearchFilter):
    def filter_queryset(self, request, queryset, view):
        search_terms = self.get_base_search_terms(request)

        # if there nothing to search - return
        if not search_terms:
            return queryset

        # first try to search in lucene way
        searchset_class = self.get_searchset_class(view, request)
        filtered_queryset = self.lucene_filter_queyset(searchset_class=searchset_class, search_terms=search_terms,
                                                       queryset=queryset)

        # if there is nothing to present
        # we give a second chance to common search
        if filtered_queryset is None:
            queryset = self.custom_filter_queryset(request=request, queryset=queryset, view=view)
        else:
            queryset = self._use_distinct(searchset_class=searchset_class,
                                          queryset=queryset, filtered_queryset=filtered_queryset)

        return queryset

    def lucene_filter_queyset(self, searchset_class, search_terms, queryset):
        if searchset_class is not None:
            try:
                return searchset_class.filter(queryset, search_terms)
            except LuceneSearchException:
                return None
        return None

    def custom_filter_queryset(self, request, queryset, view):
        return super().filter_queryset(request=request, queryset=queryset, view=view)

    def _use_distinct(self, searchset_class, queryset, filtered_queryset):
        if self.must_call_distinct(queryset, searchset_class.get_fields_sources()):
            return distinct(filtered_queryset, queryset)
        else:
            return filtered_queryset



from django.conf import settings
from django.test.signals import setting_changed


LUCYFER_SETTINGS_NAME = "LUCYFER_SETTINGS"


DEFAULTS = {
    "SAVED_SEARCHES_ENABLE": False,
    "SAVED_SEARCHES_KEY": None,
    "CACHE_SEARCH_VALUES": True,
}


class LucyferSettings:
    def __init__(self, user_settings=None, defaults=None):
        if user_settings:
            self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, LUCYFER_SETTINGS_NAME, {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError(f"Invalid setting: {attr}")

        try:
            val = self.user_settings[attr]
        except KeyError:
            val = self.defaults[attr]

        return val

    def reload(self):
        if hasattr(self, '_user_settings'):
            delattr(self, '_user_settings')


lucyfer_settings = LucyferSettings(None, DEFAULTS)


def reload_lucyfer_settings(*args, **kwargs):
    if kwargs['setting'] == LUCYFER_SETTINGS_NAME:
        lucyfer_settings.reload()


setting_changed.connect(reload_lucyfer_settings)



from collections import OrderedDict, defaultdict
from itertools import islice
from typing import Type, Dict, List, Optional

from lucyfer.settings import lucyfer_settings


class MappingValue:
    """
    Class provides interface to get search helpers for field by some prefix with cache
    """
    name: str
    sources: List[str]

    _cached_values: Optional[Dict[str, Dict[str, List[str]]]] = None
    _max_cached_values_by_prefix = 10
    _cache_values_min_length = 3

    def __init__(self, name: str,
                 field_type=None,
                 sources=None,
                 show_suggestions=True,
                 get_available_values_method=None,
                 escape_quotes_in_suggestions=True):
        self.name = name
        self.field_type = field_type
        self.sources = sources if sources else [name]
        self.show_suggestions = show_suggestions
        self.get_available_values_method = get_available_values_method
        self.escape_quotes_in_suggestions = escape_quotes_in_suggestions

    def get_values(self, qs, prefix='', cache_key=None) -> List[str]:
        if not self.show_suggestions:
            return list()

        if not lucyfer_settings.CACHE_SEARCH_VALUES:
            return self._get_parsed_values(qs=qs, prefix=prefix)

        if self._cached_values is None:
            self._cached_values = defaultdict(dict)

        if not self._cached_values[cache_key].get(prefix):
            values = self._get_parsed_values(qs=qs, prefix=prefix)

            if len(prefix) < self._cache_values_min_length:
                return values

            self._cached_values[cache_key][prefix] = values

        return self._cached_values[cache_key][prefix]

    def _get_parsed_values(self, qs, prefix):
        values = self._get_available_values(qs=qs, prefix=prefix)

        if self.escape_quotes_in_suggestions:
            values = [v.replace("'", "\\'").replace('"', '\\"') for v in values]

        return values

    def _get_available_values(self, qs, prefix):
        if callable(self.get_available_values_method):
            available_values = self.get_available_values_method()
            values = (v for v in available_values if prefix in v)
        else:
            values = self._get_values(qs, prefix)

        return list(islice(values, self._max_cached_values_by_prefix))

    def _get_values(self, qs, prefix: str) -> List[str]:
        raise NotImplementedError()


class Mapping(OrderedDict):
    """
    Mapping as a dict
    """

    _value_class: Type[MappingValue]
    _model = None

    def __init__(self, model, *args, **kwargs):
        self.model = model
        super().__init__(*args, **kwargs)

    def add_value(self, name: str,
                  field_type=None,
                  sources=None,
                  show_suggestions=True,
                  get_available_values_method=None,
                  escape_quotes_in_suggestions=True
                  ):
        if name not in self:
            self.update({name: self._value_class(name=name,
                                                 field_type=field_type,
                                                 sources=sources,
                                                 show_suggestions=show_suggestions,
                                                 get_available_values_method=get_available_values_method,
                                                 escape_quotes_in_suggestions=escape_quotes_in_suggestions
                                                 )})

    def get_values(self, qs, field_name: str, prefix: str, cache_key="DEFAULT_KEY") -> List[str]:
        """
        Get values helpers for field by prefix

        :param qs: queryset with additional filters or els search
        :param field_name: field name for get values
        :param prefix: prefix for searching values
        :param cache_key: additional argument to use possibility to cache different values for different users for ex.

        :return: list of values
        """

        if field_name not in self:
            return list()

        return self[field_name].get_values(qs=qs, prefix=prefix, cache_key=cache_key)



from typing import Dict, List

from lucyparser.tree import Operator


class BaseSearchField:
    """
    Base Field class for including in SearchSet classes
    """

    DEFAULT_LOOKUP: str
    OPERATOR_TO_LOOKUP: Dict[Operator, str] = dict()
    _default_get_available_values_method = None

    def __init__(self, sources=None,
                 exclude_sources_from_mapping=False,
                 show_suggestions=True,
                 get_available_values_method=None,
                 use_field_class_for_sources=True,
                 *args, **kwargs):

        sources = list() if sources is None else sources
        self.sources = list(set(sources))

        self.exclude_sources_from_mapping = exclude_sources_from_mapping
        self.show_suggestions = show_suggestions
        self.use_field_class_for_sources = use_field_class_for_sources
        self._get_available_values_method = get_available_values_method

    def cast_value(self, value: str):
        """
        Method for value casting if it necessary (or for ex. for search replaces)
        """
        return value

    def get_lookup(self, operator) -> str:
        """
        Returns lookup for searching
        """
        return self.OPERATOR_TO_LOOKUP.get(operator, self.DEFAULT_LOOKUP)

    def get_sources(self, field_name) -> List[str]:
        """
        Returns sources list
        """
        return self.sources or [field_name]

    def get_query(self, condition):
        """
        Returns Q object with query
        """
        raise NotImplementedError()

    def match_all(self, value) -> bool:
        """
        Check filtration necessary
        """
        return value == "*"

    def get_available_values_method(self):
        return self._get_available_values_method or self._default_get_available_values_method


def negate_query_if_necessary(func):
    """
    Decorator to invert query if conditions operator was NEQ
    """
    def wrapper(self, condition):
        query = func(self, condition)

        if query is not None and condition.operator == Operator.NEQ:
            query = query.__invert__()

        return query
    return wrapper



class LuceneSearchException(Exception):
    pass


class LuceneSearchCastValueException(LuceneSearchException):
    pass


class LuceneSearchInvalidValueException(LuceneSearchException):
    pass


def fill_field_if_it_necessary(field_name_to_constructor):
    def inner(func):
        def wrapper(*args, **kwargs):
            self_or_cls = args[0]

            for field_name, constructor in field_name_to_constructor.items():
                if getattr(self_or_cls, field_name, None) is None:
                    setattr(self_or_cls, field_name, constructor())

            return func(*args, **kwargs)
        return wrapper
    return inner



from elasticsearch_dsl import Q
from elasticsearch_dsl.query import Range
from lucyparser.tree import Operator

from lucyfer.searchset.fields.base import BaseSearchField, negate_query_if_necessary
from lucyfer.searchset.utils import FieldType
from lucyfer.utils import LuceneSearchCastValueException


class ElasticSearchField(BaseSearchField):
    DEFAULT_LOOKUP = "term"

    OPERATOR_TO_LOOKUP = {
        Operator.EQ: DEFAULT_LOOKUP,
        Operator.NEQ: DEFAULT_LOOKUP,
        Operator.GT: "gt",
        Operator.LT: "lt",
        Operator.GTE: "gte",
        Operator.LTE: "lte"
    }

    def create_query_for_sources(self, condition):
        lookup = self.get_lookup(condition.operator)
        value = self.cast_value(condition.value)
        sources = self.get_sources(condition.name)

        if lookup == self.DEFAULT_LOOKUP:
            return self._get_query_for_term(sources=sources, lookup=lookup, value=value)
        else:
            return self._get_query_for_range(sources=sources, lookup=lookup, value=value)

    def _get_query_for_term(self, sources, lookup, value):
        query = None  # if set Q() as default it will be MatchAll() anytime

        lookup = self._get_wildcard_or_lookup(value=value, lookup=lookup)

        for source in sources:
            if query is None:
                query = Q(lookup, **{source: value})
            else:
                query = query | Q(lookup, **{source: value})
        return query

    def _get_query_for_range(self, sources, lookup, value):
        query = None  # if set Q() as default it will be MatchAll() anytime
        for source in sources:
            if query is None:
                query = Range(**{source: {lookup: value}})
            else:
                query = query | Range(**{source: {lookup: value}})
        return query

    def _get_wildcard_or_lookup(self, value, lookup):
        return "wildcard" if "*" in value else lookup

    @negate_query_if_necessary
    def get_query(self, condition):
        return self.create_query_for_sources(condition=condition)


class ElasticSearchFieldWithoutWildCard(ElasticSearchField):
    def _get_wildcard_or_lookup(self, value, lookup):
        return lookup


class ElasticIntegerField(ElasticSearchFieldWithoutWildCard):
    def cast_value(self, value):
        try:
            return int(value)
        except (ValueError, TypeError):
            raise LuceneSearchCastValueException()


class ElasticFloatField(ElasticSearchFieldWithoutWildCard):
    def cast_value(self, value):
        try:
            return float(value)
        except (ValueError, TypeError):
            raise LuceneSearchCastValueException()


class ElasticBooleanField(ElasticSearchFieldWithoutWildCard):
    DEFAULT_LOOKUP = "match"

    OPERATOR_TO_LOOKUP = {
        Operator.EQ: "match",
        Operator.NEQ: "match",
    }

    _values = {"true": True, "false": False}
    _default_get_available_values_method = _values.keys

    def create_query_for_sources(self, condition):
        lookup = self.get_lookup(condition.operator)
        value = self.cast_value(condition.value)
        sources = self.get_sources(condition.name)

        if lookup == self.DEFAULT_LOOKUP:
            return self._get_query_for_term(sources=sources, lookup=lookup, value=value)
        return None

    def cast_value(self, value):
        value = value.lower()
        if value in self._values:
            return self._values[value]

        raise LuceneSearchCastValueException()


class ElasticNullBooleanField(ElasticBooleanField):
    _values = {"true": True, "false": False, "null": None}
    _default_get_available_values_method = _values.keys


default_elastic_field_types_to_fields = {
    FieldType.BOOLEAN: ElasticBooleanField,
    FieldType.INTEGER: ElasticIntegerField,
    FieldType.NULL_BOOLEAN: ElasticNullBooleanField,
    FieldType.FLOAT: ElasticFloatField,
}
